#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>

#ifndef __user
#define __user
#endif

#include "vstusb.h"

struct usb_device_descriptor {
	unsigned char  bLength;
	unsigned char  bDescriptorType;

	unsigned short bcdUSB;
	unsigned char  bDeviceClass;
	unsigned char  bDeviceSubClass;
	unsigned char  bDeviceProtocol;
	unsigned char  bMaxPacketSize0;
	unsigned short idVendor;
	unsigned short idProduct;
	unsigned short bcdDevice;
	unsigned char  iManufacturer;
	unsigned char  iProduct;
	unsigned char  iSerialNumber;
	unsigned char  bNumConfigurations;
} __attribute__ ((packed));


#define CMD_INIT				0x01
#define CMD_SET_INTEGR_TIME		0x02
#define CMD_SET_LAMP_EN			0x03
#define CMD_RESERVED			0x04
#define CMD_QUERY_INFO			0x05	//
#define CMD_WRITE_INFO			0x06	//
#define CMD_WRITE_SER_NUM	 	0x07	//
#define CMD_GET_SER_NUM	 		0x08
#define CMD_REQUEST_SPECTRA		0x09
#define CMD_SET_TRIGGER_MODE	0x0A	//
#define CMD_QUERY_NUM_PLUGINS	0x0B	//
#define CMD_QUERY_PLUGIN_IDS	0x0C	//
#define CMD_DETECT_PLUGINS		0x0D	//
#define CMD_GET_STATUS			0xfe

#define SIZEOF_CMD_INIT					0x01
#define SIZEOF_CMD_SET_INTEGR_TIME		0x03
#define SIZEOF_CMD_SET_LAMP_EN			0x03
#define SIZEOF_CMD_RESERVED				0x00
#define SIZEOF_CMD_QUERY_INFO			0x01
#define SIZEOF_CMD_WRITE_INFO			0x12
#define SIZEOF_CMD_WRITE_SER_NUM		0x11
#define SIZEOF_CMD_GET_SER_NUM	 		0x01
#define SIZEOF_CMD_REQUEST_SPECTRA		0x01
#define SIZEOF_CMD_SET_TRIGGER_MODE		0x03
#define SIZEOF_CMD_QUERY_NUM_PLUGINS	0x01
#define SIZEOF_CMD_QUERY_PLUGIN_IDS		0x01
#define SIZEOF_CMD_DETECT_PLUGINS		0x01
#define SIZEOF_CMD_GET_STATUS			0x01

#define SIZEOF_INIT					0x00
#define SIZEOF_INTEGR_TIME			0x00
#define SIZEOF_LAMP_EN				0x00
#define SIZEOF_RESERVED				0x00
#define SIZEOF_QUERY_INFO			0x12	// ep7
#define SIZEOF_WRITE_INFO			0x00
#define SIZEOF_WRITE_SER_NUM	 	0x00
#define SIZEOF_SER_NUM	 			0x11	// ep7
#define SIZEOF_SPECTRA				0x1001	// ep2 0x801???
#define SIZEOF_TRIGGER_MODE			0x00
#define SIZEOF_QUERY_NUM_PLUGINS	0x01	// ep7
#define SIZEOF_QUERY_PLUGIN_IDS		0x07	// ep7
#define SIZEOF_DETECT_PLUGINS		0x00
#define SIZEOF_STATUS				0x10	// ep7

#define OFF	0
#define ON	1

#define MODE_NORMAL	0
#define MODE_SOFTWARE	1
#define MODE_EXTERNAL	3	

#define RESET_INTEGR_TIME	100
#define RESET_A_D_FREQ		1000000
#define RESET_TRIGGER_MODE	0		// Normal Trigger
#define RESET_INTEGR_TIMER	16		// bit timer -> integration range is 3 to 65535
#define RESET_LAMP_ENABLE	OFF

#define EP2  2
#define EP7  7

#define NDR_TIMEOUT     5

struct _usb2000_state {
	int integ_time;		// integration time  - reset by initialize, set via cmd, no get
	char lamp_enable;	// reset off at init, set via cmd, no get
	char trigger_mode;	// trigger mode resets to "normal", set via cmd, no get
	int a_d_freq;		// don't know how to control...
	int integ_timer;	// don't know how to control...
} usb2000_state;

//globals
char	*TestFileName;

static int open_vstusb_dev();
static int initialize(int dev);

//static int show_device_descriptor(int dev);
static int show_integ_time(int dev);
static int show_lamp_enable(int dev);
static int show_query_info(int dev);			//
static int show_serial_number(int dev);
static int show_spectra(int dev);
static int show_trigger_mode(int dev);
static int show_query_num_plugins(int dev);		//
static int show_query_plugin_ids(int dev);		//
static int show_status(int dev);

static int set_lamp_enable(int dev, unsigned int enable);
static int set_integ_time(int dev, int int_time);
static int set_trigger_mode(int dev, int mode);
static int set_detect_plugins(int dev);						//
static int write_info(int dev, int info);					//
static int write_ser_num(int dev, int ser_num);				//

static int do_cmd_rsp(int dev, struct vstusb_args *rsp_usb_data, unsigned char *cmd, unsigned int cmdsz,
		unsigned int rspsz, unsigned int cmdpipe, unsigned int rsppipe);


int parse_cmdline (argc, argv)
{
	if (argc != 2) {
//		printf("invalid number of arguments to test program");
//		return -1;
	}

	TestFileName = NULL;

	return 0;
}

int main(int argc, char *argv[])
{
	int size;
	int i, retval=0;
	int dev = -1;			//device file pointer

	if (parse_cmdline(argc, argv) != 0)
		return -1;

	if ((dev = open_vstusb_dev()) == -1) {
		return -1;
	}

    //show_device_descriptor(dev);

    initialize(dev);

	show_status(dev);
	show_serial_number(dev);
	show_integ_time(dev);
	show_trigger_mode(dev);
	show_lamp_enable(dev);

	set_integ_time(dev, 50);
	show_integ_time(dev);
	set_trigger_mode(dev, MODE_SOFTWARE);
	show_trigger_mode(dev);

	while (retval == 0) {
		set_lamp_enable(dev, ON);
		retval = show_spectra(dev);
		set_lamp_enable(dev, OFF);
	}

	show_lamp_enable(dev);
	show_serial_number(dev);

exit:
	if (close(dev) < 0) {
		printf("%s: close err /dev/vstusb (%d)", __FUNCTION__, retval);
	}
	return 0;
}

static int clean_ep(int dev, int size, int ep)
{
	unsigned int count = 0;
	int this_read = 0;
	int retval = 0;
	int nTry = 0;
	struct vstusb_args usb_data;

	usb_data.buffer = malloc(size);
	if (!usb_data.buffer) {
		printf("clean_ep: can't malloc\n");
		return retval;
	}

	do {
		//printf("clean_ep: get %d bytes to clean ep %d\n", size-this_read, ep);

		usb_data.count = size - this_read;
		usb_data.timeout_ms = 10000;
		usb_data.pipe = ep;
		if ((retval = ioctl(dev, IOCTL_VSTUSB_RECV_PIPE, &usb_data)) != 0) {
			retval = -errno;
			//printf("clean_ep: can't recv data (%d)\n", retval);
			break;
		} else {
			//printf("clean_ep: got %d bytes from EP %d\n", usb_data.count, ep);
		}

		this_read += usb_data.count;
		//printf("clean_ep: total readings: %d bytes\n", this_read);
	} while ((this_read < size) && (++nTry < NDR_TIMEOUT));

	free(usb_data.buffer);
	return retval;
}

static int open_vstusb_dev()
{
	struct vstusb_args	usb_data;
	int dev = -1;
	int retval;
	unsigned char cmd = CMD_GET_STATUS;

	if ((dev = open("/dev/vstusb.0", O_RDWR)) == -1) {
		printf("%s: can't open /dev/vstusb.0 (%d), errno = %s\n", __FUNCTION__, dev, strerror(errno));
		return dev;
	}

	if (!(usb_data.buffer = malloc(SIZEOF_STATUS))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if ((retval = do_cmd_rsp(dev, &usb_data, &cmd, sizeof(cmd), SIZEOF_STATUS, EP2, EP7)) != 0) {
			printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
		}
		else {
			if (((char *)(usb_data.buffer))[8] != 0) {
				clean_ep(dev, SIZEOF_SPECTRA, EP2);
			}
		}
		free(usb_data.buffer);
	}
	return dev;
}

static int initialize(int dev)
{
	int retval;
	struct vstusb_args usb_data;
	unsigned char	cmd = CMD_INIT;

	usb_data.buffer = (void *)&cmd;
	usb_data.count = sizeof(cmd);
	usb_data.timeout_ms = 10000;
	usb_data.pipe = EP2;
	if ((retval = ioctl(dev, IOCTL_VSTUSB_SEND_PIPE, &usb_data)) != 0) {
		retval = -errno;
		printf("initialize: Error in initialization %d\n", retval);
	}

	usb2000_state.integ_time = RESET_INTEGR_TIME;
	usb2000_state.a_d_freq = RESET_A_D_FREQ;
	usb2000_state.trigger_mode = RESET_TRIGGER_MODE;
	usb2000_state.integ_timer = RESET_INTEGR_TIMER;
	usb2000_state.lamp_enable = RESET_LAMP_ENABLE;

	return retval;
}

#if 0
static int show_device_descriptor(int dev)
{
	int retval = -1, i;
	struct usb_device_descriptor *usb_desc;

    if (!(usb_desc = malloc(sizeof(struct usb_device_descriptor)))) {
        printf("%s: can't malloc\n", __FUNCTION__);
    }
    else {
    	if ((retval = ioctl(dev, IOCTL_VSTUSB_GET_DEVICE_DESCRIPTOR, usb_desc)) != 0) {
    		retval = -errno;
    		printf("%s: ioctl err (%d)\n", __FUNCTION__, retval);
    	}
        else {
            printf("%s: device desc (%d): ", __FUNCTION__, sizeof(struct usb_device_descriptor));
            for (i=0; i<sizeof(struct usb_device_descriptor); i++) {
                printf("%02x ", ((unsigned char *)(usb_desc))[i]);
            }
            printf("\n");
        }
    }
	return retval;
}
#endif

static int do_cmd_rsp(int dev, struct vstusb_args *rsp_usb_data, unsigned char *cmd, unsigned int cmdsz,
		unsigned int rspsz, unsigned int cmdpipe, unsigned int rsppipe)
{
	struct vstusb_args usb_data;
	int	retval = 0, i;
	// send_data(*data, sizeof(data))
	usb_data.buffer = (void *)cmd;
	usb_data.count = cmdsz;
	usb_data.timeout_ms = 10000;
	usb_data.pipe = cmdpipe;
	printf("%s: send cmd=%d sz=%d pipe=%d\n", __FUNCTION__, *cmd, cmdsz, cmdpipe);
	if ((retval = ioctl(dev, IOCTL_VSTUSB_SEND_PIPE, &usb_data)) != 0) {
		retval = -errno;
		printf("%s: can't send cmd (%d)\n", __FUNCTION__, retval);
	}
	else if (rspsz > 0) {
		rsp_usb_data->count = rspsz;
		rsp_usb_data->timeout_ms = 10000;
		rsp_usb_data->pipe = rsppipe;
		printf("%s: get rsp=%x sz=%d pipe=%d\n", __FUNCTION__, &(rsp_usb_data->buffer), rspsz, rsppipe);
		if ((retval = ioctl(dev, IOCTL_VSTUSB_RECV_PIPE, rsp_usb_data)) != 0) {
			retval = -errno;
			rsp_usb_data->count = 0;
			printf("%s: didn't get rsp (%d)\n", __FUNCTION__, retval);
		}
	}
	return retval;
}

static int show_spectra(int dev)
{
	int retval = 0, i;
	struct vstusb_args usb_data;
	unsigned char cmd = CMD_REQUEST_SPECTRA;

	if (!(usb_data.buffer = malloc(SIZEOF_SPECTRA))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		memset(usb_data.buffer, 0, SIZEOF_SPECTRA);
		if ((retval = do_cmd_rsp(dev, &usb_data, &cmd, sizeof(cmd), SIZEOF_SPECTRA, EP2, EP2)) != 0) {
			printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
		}
		else if (usb_data.count > 0) {
			printf("%s: (%d) = \n   ", __FUNCTION__, usb_data.count);
			for (i=0; i<usb_data.count; i++) {
				printf("%02x ", ((unsigned char *)usb_data.buffer)[i]);
				if (((i+1) % 64) == 0) printf("\n   ");
			}
			if (!(((i+1) % 64) == 0)) printf("\n");
		}
#if 0
		memset(usb_data.buffer, 0, SIZEOF_SPECTRA);
		if ((retval = do_cmd_rsp(dev, &usb_data, &cmd, sizeof(cmd), SIZEOF_SPECTRA, EP2, EP2)) != 0) {
			printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
		}
		else if (usb_data.count > 0) {
			printf("%s: (%d) = \n   ", __FUNCTION__, usb_data.count);
			for (i=0; i<usb_data.count; i++) {
				printf("%02x ", ((unsigned char *)usb_data.buffer)[i]);
				if (((i+1) % 64) == 0) printf("\n   ");
			}
			if (!(((i+1) % 64) == 0)) printf("\n");
		}
#endif
		free(usb_data.buffer);
	}
	return retval;
}

static int show_trigger_mode(int dev)
{
	return printf("%s: %s\n", __FUNCTION__, usb2000_state.trigger_mode == MODE_NORMAL ? 
			"Normal" : ((usb2000_state.trigger_mode == MODE_SOFTWARE) ? "Software" : "External HW"));
}

static int set_trigger_mode(int dev, int mode)
{
	int retval = -1;
	struct vstusb_args usb_data;
	unsigned char * cmd;

	if ((mode != MODE_NORMAL) && (mode != MODE_SOFTWARE) && (mode != MODE_EXTERNAL)) {
		printf("%s: invalid mode (%d)\n", __FUNCTION__, mode);
		return retval;
	}

	if (!(cmd = malloc(SIZEOF_CMD_SET_TRIGGER_MODE))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if (!(usb_data.buffer = malloc(SIZEOF_TRIGGER_MODE))) {
			printf("%s: can't malloc\n", __FUNCTION__);
		}
		else {
			cmd[0] = CMD_SET_TRIGGER_MODE;
			cmd[1] = (unsigned char)(mode);
			cmd[2] = 0;
		
			if ((retval = do_cmd_rsp(dev, &usb_data, cmd, SIZEOF_CMD_SET_TRIGGER_MODE, SIZEOF_TRIGGER_MODE, EP2, EP7)) != 0) {
				printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
			}
			else {
				usb2000_state.trigger_mode = mode;
				printf("%s: set to %d\n", __FUNCTION__, mode);
			}
			free(usb_data.buffer);
		}
		free(cmd);
	}
	return retval;
}

static int show_lamp_enable(int dev)
{
	return printf("lamp: %s\n", usb2000_state.lamp_enable == OFF ? "OFF" : "ON");
}

static int set_lamp_enable(int dev, unsigned int enable)
{
	int retval = -1;
	struct vstusb_args usb_data;
	unsigned char * cmd;

	if (!(cmd = malloc(SIZEOF_CMD_SET_LAMP_EN))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if (!(usb_data.buffer = malloc(SIZEOF_LAMP_EN))) {
			printf("%s: can't malloc\n", __FUNCTION__);
		}
		else {
			if (enable != OFF)
				enable = ON;

			cmd[0] = CMD_SET_LAMP_EN;
			cmd[1] = (unsigned char)(enable);
			cmd[2] = 0;
		
			if ((retval = do_cmd_rsp(dev, &usb_data, cmd, SIZEOF_CMD_SET_LAMP_EN, SIZEOF_LAMP_EN, EP2, EP7)) != 0) {
				printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
			}
			else {
				usb2000_state.lamp_enable = enable;
			    printf("%s: set lamp_enable to %d\n", __FUNCTION__, enable);
			}
			free(usb_data.buffer);
		}
		free(cmd);
	}
	return retval;
}

static int show_integ_time(int dev)
{
	return printf("integration time: %d\n", usb2000_state.integ_time);
}

static int set_integ_time(int dev, int int_time)
{
	int retval = -1;
	struct vstusb_args usb_data;
	unsigned char * cmd;

	if (!(cmd = malloc(SIZEOF_CMD_SET_INTEGR_TIME))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if (!(usb_data.buffer = malloc(SIZEOF_INTEGR_TIME))) {
			printf("%s: can't malloc\n", __FUNCTION__);
		}
		else {
			if (int_time < 3) {
				int_time = 3;
			}
		
			if (int_time > 65535) {
				int_time = 65535;
			}
		
			cmd[0] = CMD_SET_INTEGR_TIME;
			cmd[1] = (unsigned char) (int_time & 0xff);
			cmd[2] = (unsigned char) ((int_time >> 0x08) & 0xff);
		
			if ((retval = do_cmd_rsp(dev, &usb_data, cmd, SIZEOF_CMD_SET_INTEGR_TIME, SIZEOF_INTEGR_TIME, EP2, EP7)) != 0) {
				printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
			}
			else {
				usb2000_state.integ_time = int_time;
			    printf("%s: set integ time to %d\n", __FUNCTION__, int_time);
			}
			free(usb_data.buffer);
		}
		free(cmd);
	}
	return retval;
}

static int show_serial_number(int dev)
{
	int retval = -1;
	struct vstusb_args usb_data;
	unsigned char cmd = CMD_GET_SER_NUM;

	if (!(usb_data.buffer = malloc(SIZEOF_SER_NUM))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if ((retval = do_cmd_rsp(dev, &usb_data, &cmd, sizeof(cmd), SIZEOF_SER_NUM, EP2, EP7)) != 0) {
			printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
		}
		else if (usb_data.count > 0) {
			((char *)usb_data.buffer)[usb_data.count] = '\0';
			printf("serial number: %s\n", (char *)usb_data.buffer);
		}
		free(usb_data.buffer);
	}
	return retval;
}

static int show_status(int dev)
{
	int retval = -1, i;
	struct vstusb_args usb_data;
	unsigned char cmd = CMD_GET_STATUS;

	if (!(usb_data.buffer = malloc(SIZEOF_STATUS))) {
		printf("%s: can't malloc\n", __FUNCTION__);
	}
	else {
		if ((retval = do_cmd_rsp(dev, &usb_data, &cmd, sizeof(cmd), SIZEOF_STATUS, EP2, EP7)) != 0) {
			printf("%s: do_cmd_rsp fail (%d)\n", __FUNCTION__, retval);
		}
		else {
			printf("%s: = ", __FUNCTION__);
			for (i=0; i<usb_data.count; i++) {
				printf("%02x ", ((unsigned char *)(usb_data.buffer))[i]);
			}
			printf("\n");
		}
		free(usb_data.buffer);
	}
	return retval;
}

#if 0
#ifdef _DEBUG
    int retval = 0;
	i = 0;
	do {
	  printf("VSTUSB cleaning... status = %d", retval = get_status(vstusb));
	} while ((retval) && (++i < NDR_TIMEOUT));
#endif

#define BUF_SIZE	4096 
#define MAX_PATH    255
#define DEV_PATH    "/sys/class/usb/usb2000.0/device/"

struct operation {
    char optype;                //R/W
    char attribute[MAX_PATH];   //attribute file name
    char data[BUF_SIZE];        //data for a write
};
struct	operation op;		//next op to do

int get_test_line(t_fp)
{
    char t_line[BUF_SIZE];
    int line_idx = 0;
    int cpy_idx = 0;&count

    if (feof(t_fp)) {
        printf("end of test file\n");
        return -1;
    }

    op.attribute[0]= '\0';
    op.data[0]= '\0';
    op.optype= '\0';

    //get a line from the file
    if (fgets(t_line, 1024, t_fp) == NULL)
    {
        printf("failed to get line from test file\n");
        return -1;
    }
    printf("test line: %s", t_line);
&count
    //put it into command structure, stripping out all the junk
    if ((t_line[line_idx] != 'R') && (t_line[line_idx] != 'r') && (t_line[line_idx] != 'W') && (t_line[line_idx] != 'w')) {
        printf("malformed command line, assume it's a raw write\n");
        strcpy(op.data, t_line);
        return 0;
    }
    op.optype= t_line[line_idx++];
    
    //scan to next
    while ( (isspace(t_line[line_idx])) && (t_line[line_idx] != '\0')) {
        ++line_idx;
    }

    //get attribute name
    while ( (!isspace(t_line[line_idx])) && (cpy_idx < MAX_PATH) ) {
        op.attribute[cpy_idx++]= t_line[line_idx++];
    }
    op.attribute[cpy_idx]='\0';

    //scan to next
    while ( (isspace(t_line[line_idx])) && (t_line[line_idx] != '\0')) {
        ++line_idx;
    }

    //get data
    cpy_idx=0;
    while ( (!isspace(t_line[line_idx])) && (cpy_idx < BUF_SIZE)) {
        op.data[cpy_idx++]= t_line[line_idx++];
    }
    op.data[cpy_idx]='\0';&count

    return 0;
}

int run_test()
{
    int		a_fp;               //attribute file pointer
	FILE	*t_fp;				//test file pointer
	char	pathname[MAX_PATH];	//use for opening files

	if ((t_fp = fopen(TestFileName, "r")) < 0) {
		retval = -errno;
			printf("Failed to open test file");
			return retval;
	}
	printf("opened test file %s\n", TestFileName);

    while(!feof(t_fp))
    {
        if ((retval = get_test_line(t_fp)) != 0)
        {&count
            close_all();
            return -1;
        }

        //do the op
        //attribute file read/write&count
        if (strlen(op.attribute))
        {
            if (strcpy(pathname, DEV_PATH) != NULL)
            {
                if (strcat(pathname, op.attribute) != NULL)void close_all()
                {
                    //read
                    if (tolower(op.optype) == 'r')
                    {
                        if (a_fp = open(pathname, O_RDONLY))
                        {
                            //read the data
                            if ((size = read(a_fp, buf, sizeof(buf)-1)) <= 0)
                            {
                                printf("unable to read attribute: %s\n", pathname);
                            }
                            //terminate the string
                            buf[size] = '\0';
                            printf("read %d bytes from attribute %s, got: %s", size, pathname, buf);
                            close(a_fp);
                        }
                    }
                    //write
                    else
                    {
                        if (a_fp = open(pathname, O_RDWR))
                        {
                            //write the data
                            if ((size = write(a_fp, op.data, strlen(op.data))) <= 0)
                            {
                                printf("unable to write attribute: %s\n", pathname);
                            }
                            printf("wrote %d bytes to attribute %s, data:%s\n",size, pathname, op.data);
                            close(a_fp);
                        }
                    }
                }
            }
        }
    }
	if (t_fp != NULL) {
		if (fclose(t_fp) < 0) {
			printf("%s: close err test file (%d)", __FUNCTION__, retval);
		}
	}
}
/*
	sleep(30);

    for (loop = 0; loop < 10; loop++) {
        if ((size = read(d_fp, buf, sizeof(buf))) <= 0) {
            perror("Reading error");

            close_all();

            return 2;
        }

        printf("Good read size %d\n", size);

        for (i = 0; i < size; i++) {
	    if (isprint(buf[i])) {
                printf("c%c:", buf[i]);
	    } else {
	        printf("x%x:", buf[i]);
	    }
        }
	printf("\n");
    }
*/

#endif
