/*****************************************************************************
 *  File: drivers/usb/misc/vstusb.c
 *
 *  Purpose: Support for the bulk USB Vernier Spectrophotometers
 *
 *  Author:     Johnnie Peters
 *              Axian Consulting
 *              Beaverton, OR, USA 97005
 *
 *  Modified by:     EQware Engineering, Inc.
 *                   Oregon City, OR, USA 97045
 *
 *  Copyright:  2007, 2008
 *              Vernier Software & Technology
 *              Beaverton, OR, USA 97005
 *
 *  Web:        www.vernier.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *****************************************************************************/
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/usb.h>

#include "vstusb.h"

#define DRIVER_VERSION "VST USB Driver Version 1.4"
#define DRIVER_DESC "Vernier Software Technology Bulk USB Driver"

#ifdef CONFIG_USB_DYNAMIC_MINORS
	#define VSTUSB_MINOR_BASE	0
#else
	#define VSTUSB_MINOR_BASE	199
#endif

#define USB_VENDOR_OCEANOPTICS	0x2457
#define USB_VENDOR_VERNIER	0x08F7	/* Vernier Software & Technology */

#define USB_PRODUCT_USB2000	0x1002
#define USB_PRODUCT_ADC1000_FW	0x1003	/* firmware download (renumerates) */
#define USB_PRODUCT_ADC1000	0x1004
#define USB_PRODUCT_HR2000_FW	0x1009	/* firmware download (renumerates) */
#define USB_PRODUCT_HR2000	0x100A
#define USB_PRODUCT_HR4000_FW	0x1011	/* firmware download (renumerates) */
#define USB_PRODUCT_HR4000	0x1012
#define USB_PRODUCT_USB650	0x1014	/* "Red Tide" */
#define USB_PRODUCT_QE65000	0x1018
#define USB_PRODUCT_USB4000	0x1022
#define USB_PRODUCT_USB325	0x1024	/* "Vernier Spectrometer" */

#define USB_PRODUCT_LABPRO	0x0001
#define USB_PRODUCT_LABQUEST	0x0005

static struct usb_device_id id_table [] = {
	{ USB_DEVICE(USB_VENDOR_OCEANOPTICS, USB_PRODUCT_USB2000)},
	{ USB_DEVICE(USB_VENDOR_OCEANOPTICS, USB_PRODUCT_HR4000)},
	{ USB_DEVICE(USB_VENDOR_OCEANOPTICS, USB_PRODUCT_USB650)},
	{ USB_DEVICE(USB_VENDOR_OCEANOPTICS, USB_PRODUCT_USB4000)},
	{ USB_DEVICE(USB_VENDOR_OCEANOPTICS, USB_PRODUCT_USB325)},
	{ USB_DEVICE(USB_VENDOR_VERNIER, USB_PRODUCT_LABQUEST)},
	{ USB_DEVICE(USB_VENDOR_VERNIER, USB_PRODUCT_LABPRO)},
	{},
};

MODULE_DEVICE_TABLE(usb, id_table);

struct vstusb_device {
	struct mutex   lock;
	struct usb_device *uud_dev;
	unsigned int uud_ifnum;
	char uud_present;
	unsigned char uud_buffer[4097];
	unsigned short int uud_data[2048]; /* spectral data */
	int uud_count;			  /* data length */
	wait_queue_head_t uud_wq;	  /* for timeouts */
	char uud_isopen;
	int         rd_pipe;
	int         rd_timeout_ms;
	int         wr_pipe;
	int         wr_timeout_ms;
};

static struct usb_driver vstusb_driver;

static int vstusb_open(struct inode *inode, struct file *file)
{
	struct vstusb_device *vstdev = NULL;
	struct usb_interface *interface;

	interface = usb_find_interface(&vstusb_driver, iminor(inode));

	if (!interface) {
		err("%s - error, can't find device for minor %d\n",
			__func__, iminor(inode));
		return -ENODEV;
	}

	vstdev = usb_get_intfdata(interface);

	if (!vstdev)
		return -ENODEV;

	/* lock this device */
	if (mutex_lock_interruptible(&vstdev->lock))
		return -ERESTARTSYS;

	/* can only open one time */
	if ((!vstdev->uud_present) || (vstdev->uud_isopen)) {
		mutex_unlock(&vstdev->lock);
		return -EBUSY;
	}

	vstdev->uud_isopen = 1;

	init_waitqueue_head(&vstdev->uud_wq);

	/* save device in the file's private structure */
	file->private_data = vstdev;

	mutex_unlock(&vstdev->lock);

	dev_dbg(&vstdev->uud_dev->dev, "%s: opened\n", __func__);
	return 0;
}

static int vstusb_close(struct inode *inode, struct file *file)
{
	struct vstusb_device *vstdev = NULL;

	vstdev = file->private_data;

	if (vstdev == NULL)
		return -ENODEV;

	if (mutex_lock_interruptible(&vstdev->lock))
		return -ERESTARTSYS;

	vstdev->uud_isopen = 0;
	file->private_data = NULL;

	mutex_unlock(&vstdev->lock);

	/* if device is no longer present */
	if (!vstdev->uud_present)
		kfree(vstdev);

	return 0;
}

static void usb_api_blocking_completion(struct urb *urb, struct pt_regs *regs)
{
	complete((struct completion *)urb->context);
}

static int usbbulk_msg(struct usb_device *usb_dev, unsigned int pipe,
		       void *data, int len,
		       int *actual_length,
		       int timeout)
{
	struct urb *urb;
	struct usb_host_endpoint *ep;
	struct completion done;
	unsigned long expire;
	int status = 0;

	ep = (usb_pipein(pipe) ? usb_dev->ep_in : usb_dev->ep_out)
	     [usb_pipeendpoint(pipe)];
	if (!ep || len < 0)
		return -EINVAL;

	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb)
		return -ENOMEM;

	if ((ep->desc.bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
	    == USB_ENDPOINT_XFER_INT) {
		pipe = (pipe & ~(3 << 30)) | (PIPE_INTERRUPT << 30);
		usb_fill_int_urb(urb, usb_dev, pipe, data, len,
				 (usb_complete_t)usb_api_blocking_completion,
				 NULL, ep->desc.bInterval);
	} else
		usb_fill_bulk_urb(urb, usb_dev, pipe, data, len,
				  (usb_complete_t)usb_api_blocking_completion,
				  NULL);

	init_completion(&done);
	urb->context = &done;
	urb->actual_length = 0;
	status = usb_submit_urb(urb, GFP_NOIO);
	if (unlikely(status))
		goto exit;

	expire = timeout ? msecs_to_jiffies(timeout) : MAX_SCHEDULE_TIMEOUT;
	if (!wait_for_completion_interruptible_timeout(&done, expire)) {
		dev_dbg(&urb->dev->dev,
			"%s timed out on ep%d%s len=%d/%d, urb status = %d\n",
			current->comm,
			usb_pipeendpoint(urb->pipe),
			usb_pipein(urb->pipe) ? "in" : "out",
			urb->actual_length,
			urb->transfer_buffer_length,
			urb->status);

		usb_kill_urb(urb);
		status = urb->status == -ENOENT ? -ETIMEDOUT : urb->status;
	} else {
		if (signal_pending(current)) {
			dev_dbg(&urb->dev->dev,
				"%s: signal pending on ep%d%s len=%d/%d,"
				"urb status = %d\n",
				current->comm,
				usb_pipeendpoint(urb->pipe),
				usb_pipein(urb->pipe) ? "in" : "out",
				urb->actual_length,
				urb->transfer_buffer_length,
				urb->status);

			/* if really an error...better look at O'Reilly
			 * 3rd edition, p339 and p355 (still a valid
			 * assertion?).
			 */
			if (urb->status && !((urb->status == -ENOENT)     ||
					     (urb->status == -ECONNRESET) ||
					     (urb->status == -ESHUTDOWN))) {
				status = -EINTR;
				usb_kill_urb(urb);
			} else {
				status = 0;
			}
		} else {
			status = urb->status;
		}
	}
exit:
	if (actual_length)
		*actual_length = urb->actual_length;

	usb_free_urb(urb);
	return status;
}

static ssize_t vstusb_read(struct file *file, char __user *buffer,
			   size_t count, loff_t *ppos)
{
	struct vstusb_device *vstdev = NULL;
	int cnt = -1;
	void *buf = NULL;
	int retval = 0;
	struct usb_device       *dev;
	int                     pipe;
	int                     timeout;

	vstdev = file->private_data;

	if (vstdev == NULL)
		return -ENODEV;

	/* verify that we actually want to read some data */
	if (count == 0)
		return -EINVAL;

	/* lock this object */
	if (mutex_lock_interruptible(&vstdev->lock))
		return -ERESTARTSYS;

	/* anyone home */
	if (!vstdev->uud_present) {
		mutex_unlock(&vstdev->lock);
		err("%s: device not present\n", __func__);
		return -ENODEV;
	}

	/*pull out the necessary data*/
	dev =     vstdev->uud_dev;
	pipe =    vstdev->rd_pipe;
	timeout = vstdev->rd_timeout_ms;

	mutex_unlock(&vstdev->lock);

	buf = kmalloc(count, GFP_KERNEL);
	if (buf == NULL) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: can't malloc\n", __func__);
		return -ENOMEM;
	}

	retval = usbbulk_msg(dev, usb_rcvbulkpipe(dev, pipe), buf, count,
			     &cnt, timeout);
	if (retval) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: error %d reading from pipe %d\n",
			__func__, retval, pipe);
	} else if (copy_to_user(buffer, buf, count)) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: can't copy_to_user\n", __func__);
		retval = -EFAULT;
	} else {
		retval = cnt;
		dev_dbg(&vstdev->uud_dev->dev,
			"%s: read %d bytes from pipe %d\n",
			__func__, cnt, vstdev->rd_pipe);
	}

	kfree(buf);
	return retval;
}

static ssize_t vstusb_write(struct file *file, const char __user *buffer,
			    size_t count, loff_t *ppos)
{
	struct vstusb_device *vstdev = NULL;
	int cnt = -1;
	void *buf = NULL;
	int retval = 0;
	struct usb_device       *dev;
	int                     pipe;
	int                     timeout;

	vstdev = file->private_data;

	if (vstdev == NULL)
		return -ENODEV;

	/* verify that we actually have some data to write */
	if (count == 0)
		return retval;

	/* lock this object */
	if (mutex_lock_interruptible(&vstdev->lock))
		return -ERESTARTSYS;

	/* anyone home */
	if (!vstdev->uud_present) {
		mutex_unlock(&vstdev->lock);
		err("%s: device not present\n", __func__);
		return -ENODEV;
	}

	/*pull out the necessary data*/
	dev =     vstdev->uud_dev;
	pipe =    vstdev->wr_pipe;
	timeout = vstdev->wr_timeout_ms;

	mutex_unlock(&vstdev->lock);

	buf = kmalloc(count, GFP_KERNEL);
	if (buf == NULL) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: can't malloc\n", __func__);
		return -ENOMEM;
	}

	if (copy_from_user(buf, buffer, count)) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: can't copy_from_user\n", __func__);
		retval = -EFAULT;
	} else {
		retval = usbbulk_msg(dev, usb_sndbulkpipe(dev, pipe), buf,
				     count, &cnt, timeout);
		if (retval) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: error %d while write to pipe %d\n",
				__func__, retval, pipe);
		} else {
			retval = cnt;
			dev_dbg(&vstdev->uud_dev->dev,
				"%s: sent %d bytes to pipe %d\n",
				__func__, cnt, vstdev->wr_pipe);
		}
	}

	kfree(buf);
	return retval;
}

static int vstusb_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	int retval = 0;
	int cnt = -1;
	void __user *data = (void __user *)arg;
	struct vstusb_args usb_data;

	struct vstusb_device *vstdev = NULL;
	void *buffer = NULL;
	struct usb_device       *dev;

	vstdev = file->private_data;

	if (vstdev == NULL)
		return -ENODEV;

	if (data == NULL)
		return -EINVAL;

	if (copy_from_user(&usb_data, data, sizeof(struct vstusb_args))) {
		dev_err(&vstdev->uud_dev->dev,
			"%s: can't copy_from_user\n", __func__);
		retval = -EFAULT;
		goto exit;
	}

	/* lock this object */
	if (mutex_lock_interruptible(&vstdev->lock))
		return -ERESTARTSYS;

	/* anyone home */
	if (!vstdev->uud_present) {
		mutex_unlock(&vstdev->lock);
		err("%s: device not present\n", __func__);
		return -ENODEV;
	}
	/*pull out the necessary data*/
	dev = vstdev->uud_dev;

	mutex_unlock(&vstdev->lock);

	switch (cmd) {
	case IOCTL_VSTUSB_SEND_PIPE:
		if (usb_data.count == 0) {
			retval = -EINVAL;
			goto exit;
		}
		buffer = kmalloc(usb_data.count, GFP_KERNEL);
		if (buffer == NULL) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: can't malloc\n", __func__);
			retval = -ENOMEM;
			goto exit;
		}

		if (copy_from_user(buffer, usb_data.buffer, usb_data.count)) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: can't copy_from_user\n", __func__);
			retval = -EFAULT;
			goto exit;
		}


		retval = usbbulk_msg(dev, usb_sndbulkpipe(dev, usb_data.pipe),
				     buffer, usb_data.count, &cnt,
				     usb_data.timeout_ms);
		if (retval)
			dev_err(&vstdev->uud_dev->dev,
				"%s: error %d while send to pipe %d\n",
				__func__, retval, usb_data.pipe);
		else
			dev_dbg(&vstdev->uud_dev->dev,
				"%s: sent %d bytes to pipe %d\n",
				__func__, usb_data.count, usb_data.pipe);
		break;

	case IOCTL_VSTUSB_RECV_PIPE:
		if (usb_data.count == 0) {
			retval = -EINVAL;
			goto exit;
		}
		buffer = kmalloc(usb_data.count, GFP_KERNEL);
		if (buffer == NULL) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: can't malloc\n", __func__);
			retval = -ENOMEM;
			goto exit;
		}
		retval = usbbulk_msg(dev, usb_rcvbulkpipe(dev, usb_data.pipe),
				     buffer, usb_data.count, &cnt,
				     usb_data.timeout_ms);
		if (retval) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: error %d receiving from pipe %d\n",
				 __func__, retval, usb_data.pipe);
		} else if (copy_to_user(usb_data.buffer, buffer, cnt)) {
			dev_err(&vstdev->uud_dev->dev,
				"%s: can't copy_to_user\n", __func__);
			retval = -EFAULT;
		} else {
			usb_data.count = cnt;
			if (copy_to_user(data, 
					 &usb_data, 
					 sizeof(struct vstusb_args))) {
				dev_err(&vstdev->uud_dev->dev,
					"%s: can't copy_to_user\n", __func__);
				retval = -EFAULT;
			} else {
				dev_dbg(&vstdev->uud_dev->dev,
					"%s: recv %d bytes from pipe %d\n",
					__func__,
					usb_data.count,
					usb_data.pipe);
			}
		}
		break;

	case IOCTL_VSTUSB_CONFIG_RW:
		/* lock this object */
		if (mutex_lock_interruptible(&vstdev->lock))
			return -ERESTARTSYS;

		vstdev->rd_pipe = usb_data.rd_pipe;
		vstdev->rd_timeout_ms = usb_data.rd_timeout_ms;
		vstdev->wr_pipe = usb_data.wr_pipe;
		vstdev->wr_timeout_ms = usb_data.wr_timeout_ms;

		mutex_unlock(&vstdev->lock);

		dev_dbg(&vstdev->uud_dev->dev,
			"%s: setting pipes/timeouts, "
			"rdpipe = %d, rdtimeout = %d, "
			"wrpipe = %d, wrtimeout = %d\n", __func__,
			vstdev->rd_pipe, vstdev->rd_timeout_ms,
			vstdev->wr_pipe, vstdev->wr_timeout_ms);
		break;

	default:
		dev_warn(&vstdev->uud_dev->dev,
			 "ioctl_vstusb: invalid ioctl cmd %x\n", cmd);
		return -EINVAL;
		break;
	}
exit:
	kfree(buffer);
	return retval;
}

static const struct file_operations vstusb_fops = {
	.owner =    THIS_MODULE,
	.read =     vstusb_read,
	.write =    vstusb_write,
	.ioctl =    vstusb_ioctl,
	.open =     vstusb_open,
	.release =  vstusb_close,
};

static struct usb_class_driver usb_vstusb_class = {
	.name =         "usb/vstusb%d",
	.fops =         &vstusb_fops,
	.minor_base =   VSTUSB_MINOR_BASE,
};

static int vstusb_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct vstusb_device *vstdev = NULL;
	int i;
	int retval = 0;

	/* allocate memory for our device state and intialize it */

	vstdev = kzalloc(sizeof(*vstdev), GFP_KERNEL);
	if (vstdev == NULL) {
		dev_err(&intf->dev, "Out of memory\n");
		return -ENOMEM;
	}

	mutex_init(&vstdev->lock);

	i = dev->descriptor.bcdDevice;

	dev_info(&intf->dev,
		 "%s: Version %1d%1d.%1d%1d found at address %d\n", __func__,
		 (i & 0xF000) >> 12, (i & 0xF00) >> 8,
		 (i & 0xF0) >> 4, (i & 0xF),
		 dev->devnum);

	vstdev->uud_present = 1;
	vstdev->uud_isopen = 0;
	vstdev->uud_dev = dev;

	retval = usb_register_dev(intf, &usb_vstusb_class);
	if (retval) {
		dev_err(&intf->dev,
			"%s: Not able to get a minor for this device.\n",
			__func__);
		kfree(vstdev);
		return retval;
	}

	usb_set_intfdata(intf, vstdev);

	/* let the user know what node this device is now attached to */
	dev_info(&intf->dev,
		 "VST USB Device #%d now attached to major %d minor %d\n",
		 (intf->minor - VSTUSB_MINOR_BASE), USB_MAJOR, intf->minor);

	dev_info(&intf->dev, "%s, %s\n", DRIVER_DESC, DRIVER_VERSION);

	return retval;
}

static void vstusb_disconnect(struct usb_interface *intf)
{
	struct vstusb_device *vstdev = usb_get_intfdata(intf);

	usb_set_intfdata(intf, NULL);
	usb_deregister_dev(intf, &usb_vstusb_class);

	if (vstdev) {

		mutex_lock(&vstdev->lock);
		vstdev->uud_present = 0;
		mutex_unlock(&vstdev->lock);

		/* if the device is not opened, then we clean up right now */
		if (!vstdev->uud_isopen) {
			kfree(vstdev);
		}
	}
}

static struct usb_driver vstusb_driver = {
	.name =         "vstusb",
	.probe =        vstusb_probe,
	.disconnect =   vstusb_disconnect,
	.id_table = id_table,
};

static int __init vstusb_init(void)
{
	int rc;

	rc = usb_register(&vstusb_driver);
// info call is NOT supported on recent (~2.6.31-16 generic) kernels
//	if (rc)
//		info("%s: failed to register (%d)", __func__, rc);
//	else
//		info("%s: registered", __func__);

	return rc;
}

static void __exit vstusb_exit(void){
	usb_deregister(&vstusb_driver);
//	info("%s: deregistered", __func__);
}

module_init(vstusb_init);
module_exit(vstusb_exit);

MODULE_AUTHOR("Dennis O'Brien/Stephen Ware");
MODULE_DESCRIPTION(DRIVER_VERSION);
MODULE_LICENSE("GPL");
